package xerror

import (
	"fmt"

	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xcode"
)

// New 用于创建一个自定义文本错误信息的 error 对象，并包含堆栈信息。
func New(text string) error {
	return &Error{
		stack: callers(),
		text:  text,
		code:  xcode.CodeDefault,
	}
}

// NewC add code
func NewC(code int, text string) error {
	return &Error{
		stack: callers(),
		text:  text,
		code:  code,
	}
}

func Join(errs ...error) error {
	if len(errs) == 0 {
		return nil
	}
	var es []error
	for _, e := range errs {
		if e != nil {
			es = append(es, e)
		}
	}
	if len(es) == 0 {
		return nil
	}
	if len(es) == 1 {
		return es[0]
	}
	err := es[0]
	for i := 1; i < len(es); i++ {
		err = Wrap(err, es[i].Error())
	}
	return err
}

// Newf 用于创建一个自定义文本错误带参数信息的 error 对象，并包含堆栈信息。
func Newf(format string, args ...any) error {
	return &Error{
		stack: callers(),
		text:  fmt.Sprintf(format, args...),
		code:  xcode.CodeDefault,
	}
}
func Errorf(format string, args ...any) error {
	return &Error{
		stack: callers(),
		text:  fmt.Sprintf(format, args...),
		code:  xcode.CodeDefault,
	}
}
func NewCf(code int, format string, args ...any) error {
	return &Error{
		stack: callers(),
		text:  fmt.Sprintf(format, args...),
		code:  code,
	}
}

// NewSkip 用于创建一个自定义错误信息的 error 对象，并且忽略部分堆栈信息（按照当前调用方法位置往上忽略）。高级功能，一般开发者很少用得到。
// 参数 `skip` 指定堆栈跳过的层数。
func NewSkip(skip int, text string) error {
	return &Error{
		stack: callers(skip),
		text:  text,
		code:  xcode.CodeDefault,
	}
}

func NewSkipC(code, skip int, text string) error {
	return &Error{
		stack: callers(skip),
		text:  text,
		code:  code,
	}
}

// NewSkipf 用于创建一个自定义错误信息的error对象，并且忽略部分堆栈信息（按照当前调用方法位置往上忽略）。
// 参数 `skip` 指定堆栈跳过的层数。
func NewSkipf(skip int, format string, args ...any) error {
	return &Error{
		stack: callers(skip),
		text:  fmt.Sprintf(format, args...),
		code:  xcode.CodeDefault,
	}
}

func NewSkipCf(code, skip int, format string, args ...any) error {
	return &Error{
		stack: callers(skip),
		text:  fmt.Sprintf(format, args...),
		code:  code,
	}
}

// Wrap 用于包裹其他错误 error 对象，构造成多级的错误信息，包含堆栈信息。
// 注：它不会丢失包装错误的错误码，因为它从中继承了错误码。
func Wrap(err error, text string) error {
	if err == nil {
		return nil
	}
	return &Error{
		error: err,
		stack: callers(),
		text:  text,
		code:  Code(err),
	}
}

// Wrapf 用于包裹其他错误 error 对象，构造成多级的错误信息，包含堆栈信息。
func Wrapf(err error, format string, args ...any) error {
	if err == nil {
		return nil
	}
	return &Error{
		error: err,
		stack: callers(),
		text:  fmt.Sprintf(format, args...),
		code:  Code(err),
	}
}

// WrapSkip 用于包裹其他错误 error 对象，构造成多级的错误信息，包含堆栈信息，并且忽略部分堆栈信息（按照当前调用方法位置往上忽略）。
func WrapSkip(skip int, err error, text string) error {
	if err == nil {
		return nil
	}
	return &Error{
		error: err,
		stack: callers(skip),
		text:  text,
		code:  Code(err),
	}
}

// WrapSkipf 用于包裹其他错误 error 对象，构造成多级的错误信息，包含堆栈信息，并且忽略部分堆栈信息（按照当前调用方法位置往上忽略）。
func WrapSkipf(skip int, err error, format string, args ...any) error {
	if err == nil {
		return nil
	}
	return &Error{
		error: err,
		stack: callers(skip),
		text:  fmt.Sprintf(format, args...),
		code:  Code(err),
	}
}

type CodeI interface {
	Code() int
}

func GetCodeOrDefault(err error, defaultCode int) int {
	if err == nil {
		return defaultCode
	}
	if code, ok := err.(CodeI); ok {
		return code.Code()
	}
	return defaultCode
}
